
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <Mowitz/MwUtils.h>
#include "MwHtmlParser.h"
#include "kylie.h"

struct loader {
	char *prot;
	url_info *(*load)(char *);
};

static struct loader loaders[] = {
	{"http", load_http},
	{"file", load_file},
	{"about", load_about},
	{NULL, NULL}};

#if 0
/* Change // to /, /./ to / and /xxx/../ to / */
static void delint_url(char *p)
{
        char *q, *r;

        while (*p && isalpha(*p)) p++;
        if (*p++ != ':' || *p++ != '/' || *p++ != '/') return;
        while (*p && *p != '/') p++;
        /* we're now at the first / of the uri */

        /* change // to / */
        while ((q = strstr(p, "//")))
                memmove(q, q+1, strlen(q+1)+1);

        /* change /./ to / */
        while ((q = strstr(p, "/./")))
                memmove(q, q+2, strlen(q+2)+1);

        /* change /xxx/../ to / */
        while ((q = strstr(p, "/../"))) {
                r = q-1;
                while (r >= p && *r != '/') r--;
                if (*r != '/') break;
                memmove(r, q+3, strlen(q+3)+1);
        }
}
#endif

#if 0	/* we only use already-normalized urls */
/*
Returns a newly allocated, normalized url. Caller must free.
protocol://host/path/file.ext

The behaviour is this:

If we type something into the location box, it is taken literally,
so "from" is empty. We don't know if the url is complete or not, so
we must check that. Examples:                                      
/foo/bar should be file:///foo/bar
foo/bar should be file:///`pwd`/foo/bar if there is such a file (check
with stat), otherwise http://foo/bar. Note that IE botches this: when 
it sees host:port/whatever, the host is interpreted as protocol.     
*/
static char *normalize_url(char *url)
{
        char b[1024], c[1024];
        struct stat statbuf;

        /* skip leading space */
        while (isspace(*url)) url++;

        if (url[0] == '/') {
                /* /foo/bar should be file:///foo/bar */
                sprintf(b, "file://%s", url);
        } else if (!strstr(url, "://")) {
                /* if it does *not* contain "://", try local file */
                getcwd(c, sizeof c);
                strcat(c, "/");
                strcat(c, "url");
                if (!stat(c, &statbuf)) {
                        sprintf(b, "file://%s", c);
                } else {
                        strcpy(b, url);
                }
        } else {
                strcpy(b, url);
        }

        /* OK, now it is a more or less complete url */
        delint_url(b);
        return MwStrdup(b);
}
#endif

/* Reference counted cache */
struct url_cache {
	url_info *ui;
	int ref;
	struct url_cache *next;
} *uc;

void free_url(url_info *u1)
{
	struct url_cache *u;

	if (u1 == NULL) return;

	for (u = uc; u->ui != u1; u = u->next);
	if (u) {
		u->ref--;
		return;		/* don't free right now */
	}
}

/* Input state: ui->base = url of parent doc or NULL,
   ui->url = requested url, ui->state = current state.
   Valid states are -1 = error, 0 = done, 1 = beginning (as above),
   2 = ui->nurl contains normalized url,
   3 = ui->local contains local file name,
   4 = has been redirected once, 5 = has been redirected twice.
*/
url_info *load_url(char *url)
{
	int i;
        char *p;
	struct url_cache *u, *u1;
	url_info *ui = NULL;

	for (u = uc; u; u = u->next) {
		if (!strcmp(url, u->ui->url)) {
			u->ref++;
			return u->ui;
		}
	}
	/* and now it's all up to the protocol handlers */
	p = strchr(url, ':');
	if (p) {
		for (i = 0; loaders[i].prot; i++) {
			char pr[10];
			int n = p-url;
			if (n > 9) n = 9;
			strncpy(pr, url, n);
			pr[n] = '\0';
			if (!MwStrcasecmp(loaders[i].prot, pr)) {
				break;
			}
		}
	}
	if (p && loaders[i].prot) {
		ui = (*loaders[i].load)(url);
	}
	if (ui == NULL) return NULL;
	u1 = MwMalloc(sizeof *u1);
	u1->ui = ui;
	u1->ref = 1;
	u1->next = uc;
	uc = u1;
	return ui;
}

